/*
 * Copyright 2012 Red Hat Inc.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE COPYRIGHT HOLDER(S) OR AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Authors: Ben Skeggs
 */

#include "drmP.h"

#include "nouveau_drv.h"
#include "nouveau_ramht.h"
#include "nouveau_software.h"
#include "nouveau_gpuobj.h"

#include "nv50_display.h"

struct nvc0_software_priv {
	struct nouveau_software_priv base;
};

struct nvc0_software_chan {
	struct nouveau_software_chan base;
	struct nouveau_vma dispc_vma[4];
};

u64
nvc0_software_crtc(struct nouveau_channel *chan, int crtc)
{
	struct nvc0_software_chan *pch = chan->engctx[NVDEV_ENGINE_SW];
	return pch->dispc_vma[crtc].offset;
}

static int
nvc0_software_context_new(struct nouveau_channel *chan, int engine)
{
	struct nouveau_device *ndev = chan->device;
	struct nvc0_software_priv *psw = nv_engine(ndev, NVDEV_ENGINE_SW);
	struct nvc0_software_chan *pch;
	int ret = 0, i;

	pch = kzalloc(sizeof(*pch), GFP_KERNEL);
	if (!pch)
		return -ENOMEM;

	nouveau_software_context_new(&pch->base);
	chan->engctx[engine] = pch;

	/* map display semaphore buffers into channel's vm */
	for (i = 0; !ret && i < ndev->dev->mode_config.num_crtc; i++) {
		struct nouveau_bo *bo;
		if (ndev->card_type >= NV_D0)
			bo = nvd0_display_crtc_sema(ndev, i);
		else
			bo = nv50_display(ndev)->crtc[i].sem.bo;

		ret = nouveau_bo_vma_add(bo, chan->vm, &pch->dispc_vma[i]);
	}

	if (ret)
		psw->base.base.context_del(chan, engine);
	return ret;
}

static void
nvc0_software_context_del(struct nouveau_channel *chan, int engine)
{
	struct nouveau_device *ndev = chan->device;
	struct nvc0_software_chan *pch = chan->engctx[engine];
	int i;

	if (ndev->card_type >= NV_D0) {
		for (i = 0; i < ndev->dev->mode_config.num_crtc; i++) {
			struct nouveau_bo *bo = nvd0_display_crtc_sema(ndev, i);
			nouveau_bo_vma_del(bo, &pch->dispc_vma[i]);
		}
	} else
	if (ndev->card_type >= NV_50) {
		struct nv50_display *disp = nv50_display(ndev);
		for (i = 0; i < ndev->dev->mode_config.num_crtc; i++) {
			struct nv50_display_crtc *dispc = &disp->crtc[i];
			nouveau_bo_vma_del(dispc->sem.bo, &pch->dispc_vma[i]);
		}
	}

	chan->engctx[engine] = NULL;
	kfree(pch);
}

static int
nvc0_software_object_new(struct nouveau_channel *chan, int engine,
			 u32 handle, u16 class)
{
	return 0;
}

int
nvc0_software_create(struct nouveau_device *ndev, int engine)
{
	struct nvc0_software_priv *priv;
	int ret;

	ret = nouveau_engine_create(ndev, engine, "SW", "software", &priv);
	if (ret)
		return ret;

	priv->base.base.context_new = nvc0_software_context_new;
	priv->base.base.context_del = nvc0_software_context_del;
	priv->base.base.object_new = nvc0_software_object_new;
	nouveau_software_create(&priv->base);

	NVOBJ_CLASS(ndev, 0x906e, SW);
	return nouveau_engine_init(ndev, engine, ret);
}
